#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
from contextlib import contextmanager
from typing import Generator, Optional

from training_telemetry.metrics import CheckpointMetrics, EventMetrics
from training_telemetry.provider import Provider
from training_telemetry.recorder import Recorder
from training_telemetry.spans import Span, SpanColor, SpanName
from training_telemetry.verbosity import Verbosity


def get_recorder() -> Recorder:
    """Shortcut for getting the recorder from the provider."""
    return Provider.instance().recorder


@contextmanager
def timed_span(
    name: SpanName,
    color: Optional[SpanColor] = None,
    start_time: Optional[float] = None,
    verbosity: Verbosity = Verbosity.INFO,
    metrics: Optional[EventMetrics] = None,
) -> Generator[Span, None, None]:
    """Context manager for recording a span with timing and metrics.
    The initial set of metrics will be recorded with the start event. If the user adds more metrics to the span,
    they will be recorded with the end event.

    Args:
        name: The name/type of the span to record
        start_time: The start time of the span
        color: The color to use for the span
        verbosity: The verbosity level to use for recording the span
        metrics: Optional dictionary of metrics to record with the start event

    Yields:
        the span
    """
    recorder = get_recorder()
    span = recorder.start(name=name, color=color, start_time=start_time, verbosity=verbosity, metrics=metrics)

    try:
        yield span
    except Exception as e:
        recorder.error(f"Error in {name}:", e, span)
        raise e
    finally:
        recorder.stop(span)


@contextmanager
def running(
    start_time: Optional[float] = None,
    metrics: Optional[EventMetrics] = None,
    verbosity: Verbosity = Verbosity.INFO,
    color: Optional[SpanColor] = None,
) -> Generator[Span, None, None]:
    """
    Context manager for recording events for the entire application duration.
    This context or function annotation can be used for the main function.
    This will result in any exceptions that occur and that are not caught by the application
    to be recorded as an error event.

    Args:
        metrics: Optional dictionary of metrics to record with the start event
        verbosity: The verbosity level to use for recording the span
        color: The color to use for the span

    Yields:
        the span
    """
    with timed_span(
        name=SpanName.MAIN_FUNCTION,
        color=color,
        start_time=start_time,
        verbosity=verbosity,
        metrics=metrics,
    ) as span:
        yield span


@contextmanager
def training(
    start_time: Optional[float] = None,
    metrics: Optional[EventMetrics] = None,
    verbosity: Verbosity = Verbosity.INFO,
    color: Optional[SpanColor] = None,
) -> Generator[Span, None, None]:
    """Context manager for recording the entire training loop.
    This is just a wrapper of timed_span() with the event name set to TRAINING_LOOP.

    Args:
        metrics: Optional dictionary of metrics to record with the start event
        verbosity: The verbosity level to use for recording the span
        color: The color to use for the span

    Yields:
        the span
    """
    with timed_span(
        name=SpanName.TRAINING_LOOP,
        color=color,
        start_time=start_time,
        verbosity=verbosity,
        metrics=metrics,
    ) as span:
        yield span


@contextmanager
def validation(
    start_time: Optional[float] = None,
    metrics: Optional[EventMetrics] = None,
    verbosity: Verbosity = Verbosity.INFO,
    color: Optional[SpanColor] = None,
) -> Generator[Span, None, None]:
    """Context manager for recording the entire validation loop.
    This is just a wrapper of timed_span() with the event name set to VALIDATION_RUNNING.

    Args:
        metrics: Optional dictionary of metrics to record with the start event
        verbosity: The verbosity level to use for recording the span
        color: The color to use for the span

    Yields:
        the span
    """
    with timed_span(
        name=SpanName.VALIDATION_LOOP,
        color=color,
        start_time=start_time,
        verbosity=verbosity,
        metrics=metrics,
    ) as span:
        yield span


@contextmanager
def checkpoint_save(
    start_time: Optional[float] = None,
    metrics: Optional[CheckpointMetrics] = None,
    verbosity: Verbosity = Verbosity.INFO,
    color: Optional[SpanColor] = None,
) -> Generator[Span, None, None]:
    """
    Context manager for recording a checkpoint save with timing and metrics. This is just
    a wrapper of timed_span() with the event name set to CHECKPOINT_SAVE.

    Args:
        metrics: Optional dictionary of metrics to record with the start event
        verbosity: The verbosity level to use for recording the span
        color: The color to use for the span

    Yields:
        the span
    """
    with timed_span(
        name=SpanName.CHECKPOINT_SAVE_SYNC,
        color=color,
        start_time=start_time,
        verbosity=verbosity,
        metrics=metrics,
    ) as span:
        yield span



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import json
import logging
from typing import Any

from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor, LogExporter
from opentelemetry.sdk.resources import Resource

from training_telemetry.config import ApplicationConfig, BackendType, OTLPLogsFormat
from training_telemetry.internal.backends.event_based_backend import EventBasedBackend, RecordWriter
from training_telemetry.internal.backends.logger_backend import PythonLoggerRecordWriter
from training_telemetry.verbosity import Verbosity


def make_otlp_logger(
    name: str, level: int, exporter: LogExporter, application: ApplicationConfig, version: str
) -> tuple[LoggerProvider, logging.Logger]:
    resource: Resource = Resource.create(
        {
            "job_name": application.job_name,
            "job_id": application.job_id,
            "job_restart_count": application.job_restart_count,
            "environment": application.environment,
            "rank": application.rank,
            "world_size": application.world_size,
            "training_telemetry_version": version,
        }
    )
    provider = LoggerProvider(resource=resource)
    processor = BatchLogRecordProcessor(exporter)
    provider.add_log_record_processor(processor)

    # Create a Python logger that will send logs through OpenTelemetry
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # Create a LoggingHandler that will send logs to OpenTelemetry
    handler = LoggingHandler(level=level, logger_provider=provider)
    logger.addHandler(handler)

    # Prevent propagation to avoid duplicate logs
    logger.propagate = False

    return provider, logger


class OTLPythonLogsRecordWriter(PythonLoggerRecordWriter):
    """Backend implementation that sends OpenTelemetry logs over OTLP."""

    def __init__(
        self,
        logger: logging.Logger,
        provider: LoggerProvider,
    ) -> None:
        self._provider = provider
        self._logger = logger
        super().__init__(logger)

    def close(self) -> None:
        super().close()
        self._provider.shutdown()


class OTLJsonLogsRecordWriter(RecordWriter):
    """Backend implementation that sends OpenTelemetry logs over OTLP."""

    def __init__(
        self,
        logger: logging.Logger,
        provider: LoggerProvider,
    ) -> None:
        self._provider = provider
        self._logger = logger

    def serialize(self, record: dict[str, Any]) -> str:
        return json.dumps(record)

    def write(self, record: str) -> None:
        self._logger.info(record)

    def close(self) -> None:
        self._provider.shutdown()


class OTLPLogsBackend(EventBasedBackend):
    """Backend implementation that sends OpenTelemetry logs over OTLP."""

    def __init__(
        self,
        name: str,
        level: int,
        exporter: LogExporter,
        verbosity: Verbosity,
        application: ApplicationConfig,
        version: str,
        format: OTLPLogsFormat,
    ) -> None:
        self._verbosity = verbosity
        self._provider, self._logger = make_otlp_logger(name, level, exporter, application, version)
        self._record_writer = (
            OTLPythonLogsRecordWriter(self._logger, self._provider)
            if format == OTLPLogsFormat.KEY_VALUE
            else OTLJsonLogsRecordWriter(self._logger, self._provider)
        )
        super().__init__(self._record_writer)

    def verbosity(self) -> Verbosity:
        return self._verbosity

    def type(self) -> BackendType:
        return BackendType.OTLP_LOGS

    def close(self) -> None:
        super().close()



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
from abc import ABC, abstractmethod
from typing import Optional

from training_telemetry.events import Event, ExceptionEvent
from training_telemetry.metrics import EventMetrics
from training_telemetry.spans import Span, SpanColor, SpanName
from training_telemetry.verbosity import Verbosity


class Recorder(ABC):
    """An abstract base class for creating and recording spans and events."""

    @abstractmethod
    def start(
        self,
        name: SpanName,
        color: Optional[SpanColor] = None,
        start_time: Optional[float] = None,
        verbosity: Verbosity = Verbosity.INFO,
        metrics: Optional[EventMetrics] = None,
    ) -> Span:
        """
        Create a new span and record its start event.

        Args:
            name: The name of the span.
            color: The optional color of the span.
            start_time: The optional start time of the span, the current timestamp will be used if not specified.
            verbosity: The verbosity of the span, INFO by default.
            metrics: The optional metrics of the span, the start event will have these metrics added to it.
        """
        pass

    @abstractmethod
    def stop(self, span: Span) -> None:
        """
        Stop the span and record its end event. The same verbosity level as the start event is used.
        Args:
            span: The span to stop.
        """
        pass

    @abstractmethod
    def event(self, event: Event, span: Optional[Span] = None, verbosity: Verbosity = Verbosity.INFO) -> Event:
        """
        Create a new event and record it immediately.
        The event is added to the last span that was started, unless a span is provided.

        Args:
            event: The event to record.
            span: Optional span to add the event to, the last started span will be used if not specified.
            verbosity: The verbosity level of the event, INFO by default.
        Returns:
            The event recorded.
        """
        pass

    @abstractmethod
    def error(
        self,
        error_message: str,
        exception: Optional[Exception] = None,
        span: Optional[Span] = None,
        verbosity: Verbosity = Verbosity.ERROR,
    ) -> ExceptionEvent:
        """
        Create a new error event with the given error message and exception and record it immediately.
        The error event is added to the last span that was started, unless a span is provided.

        Args:
            error_message: The message of the error event.
            exception: The exception of the error event.
            span: Optional span to add the error event to, the last started span will be used if not specified.
            verbosity: The verbosity level of the error event, ERROR by default.
        Returns:
            The error event recorded.
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Close the recorder and release any resources.
        All spans that were previously started will be ended and thei end events recorded at the current time.
        Once the recorder is closed, it cannot be used again.
        """
        pass



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
__version__ = "1.1.0"



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import traceback
import uuid
from collections import OrderedDict
from typing import Optional

from training_telemetry.events import Event, ExceptionEvent
from training_telemetry.internal.backend import Backend
from training_telemetry.metrics import EventMetrics
from training_telemetry.recorder import Recorder
from training_telemetry.spans import Span, SpanColor, SpanName
from training_telemetry.utils import get_default_logger
from training_telemetry.verbosity import Verbosity

_logger = get_default_logger()


class StandardRecorder(Recorder):
    """
    Standard implementation of the Recorder interface supporting multiple backends and tracking of spans.
    It is possible to create a StandardRecorder with no backends, in which case it will not record anything,
    this is useful if the application is running on thousands of GPUs, in which case the overhead of recording
    from each GPU is too high, and we only want to record from rank zero or a subset of ranks.
    """

    def __init__(self, backends: list[Backend]):
        self._backends = backends.copy()
        self._spans: OrderedDict[uuid.UUID, Span] = OrderedDict()
        if len(self._backends) > 0:
            _logger.debug(f"Initializing StandardRecorder with {len(self._backends)} backends")
        else:
            _logger.debug("Initializing StandardRecorder with no backends, recording is disabled")

    @property
    def backends(self) -> list[Backend]:
        return self._backends

    def start(
        self,
        name: SpanName,
        color: Optional[SpanColor] = None,
        start_time: Optional[float] = None,
        verbosity: Verbosity = Verbosity.INFO,
        metrics: Optional[EventMetrics] = None,
    ) -> Span:
        span = Span.create(name=name, color=color, start_time=start_time, verbosity=verbosity, metrics=metrics)
        self._spans[span.id] = span
        for backend in self._backends:
            try:
                if verbosity <= backend.verbosity():
                    backend.record_start(span)
            except Exception as e:
                _logger.error(f"Error recording start for span {span.id}: {e}")
        return span

    def stop(self, span: Span) -> None:
        if span.id not in self._spans:
            raise ValueError(f"Span {span.id} not found")

        del self._spans[span.id]
        span.stop()
        for backend in self._backends:
            try:
                if span.verbosity <= backend.verbosity():
                    backend.record_stop(span)
            except Exception as e:
                _logger.error(f"Error recording stop for span {span.id}: {e}")

    def event(self, event: Event, span: Optional[Span] = None, verbosity: Verbosity = Verbosity.INFO) -> Event:
        if span is None:
            if self._spans:
                span = next(reversed(self._spans.values()))
            else:
                raise ValueError("No span to add event to")
        else:
            if span.id not in self._spans:
                raise ValueError(f"Span {span.id} not found")

        span.add_event(event)
        for backend in self._backends:
            try:
                if verbosity <= backend.verbosity():
                    backend.record_event(event, span)
            except Exception as e:
                _logger.error(f"Error recording event {event.name} for span {span.id}: {e}")
        return event

    def error(
        self,
        error_message: str,
        exception: Optional[Exception] = None,
        span: Optional[Span] = None,
        verbosity: Verbosity = Verbosity.ERROR,
    ) -> ExceptionEvent:
        if span is None:
            if self._spans:
                span = next(reversed(self._spans.values()))
            else:
                raise ValueError("No span to add event to")
        else:
            if span.id not in self._spans:
                raise ValueError(f"Span {span.id} not found")

        event = ExceptionEvent.create(
            error_message=error_message, exception=exception, exception_traceback=traceback.format_exc()
        )
        span.add_event(event)
        for backend in self._backends:
            try:
                if verbosity <= backend.verbosity():
                    backend.record_error(event, span)
            except Exception as e:
                _logger.error(f"Error recording error {error_message} for span {span.id}: {e}")
        return event

    def close(self) -> None:
        try:
            for span in reversed(self._spans.values()):
                for backend in self._backends:
                    try:
                        if span.verbosity <= backend.verbosity():
                            backend.record_stop(span)
                    except Exception as e:
                        _logger.error(f"Error recording stop for span {span.id}: {e}")
        finally:
            self._spans.clear()



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import uuid
from enum import Enum
from typing import Optional

from training_telemetry.duration import Duration
from training_telemetry.events import Event, EventName
from training_telemetry.metrics import EventMetric, EventMetrics
from training_telemetry.verbosity import Verbosity


class SpanName(str, Enum):
    """
    Enum class representing names of predefined spans.
    """

    ##
    ## Long running spans
    ##

    # The entire duration of the main or launcher function of the application.
    MAIN_FUNCTION = "main_function"
    # The entire duration of the training loop.
    TRAINING_LOOP = "training_loop"
    # The entire duration of a validation phase.
    VALIDATION_LOOP = "validation_loop"

    ##
    ## Initialization spans
    ##

    # The initialization of the distributed training code.
    DIST_INIT = "distributed_code_initialization"
    # The initialization of the data loader.
    DATA_LOADER_INIT = "data_loader_initialization"
    # The initialization of the model.
    MODEL_INIT = "model_initialization"
    # The initialization of the optimizer.
    OPTIMIZER_INIT = "optimizer_initialization"

    ##
    ## Checkpoint spans
    ##

    # The loading of a checkpoint.
    CHECKPOINT_LOAD = "checkpoint_load"
    # The saving of a checkpoint synchronously.
    CHECKPOINT_SAVE_SYNC = "checkpoint_save_sync"
    # The saving of a checkpoint asynchronously.
    # Only report exposed checkpoint time, i.e. the time the training loop was blocked for.
    CHECKPOINT_SAVE_ASYNC = "checkpoint_save_async"
    # The finalization of an asynchronous checkpoint save.
    CHECKPOINT_SAVE_FINALIZATION = "checkpoint_save_finalization"

    ##
    ## Very short spans, normally used for profiling backends
    ##

    # A single iteration in either the training or validation loop, also known as a single step.
    ITERATION = "iteration"
    # The loading of data inside a training iteration.
    DATA_LOADING = "data_loading"
    # The forward pass inside a training iteration.
    MODEL_FORWARD = "model_forward"
    # The zeroing of the gradients.
    ZERO_GRAD = "zero_grad"
    # The backward pass inside a training iteration.
    MODEL_BACKWARD = "model_backward"
    # The update of the optimizer.
    OPTIMIZER_UPDATE = "optimizer_update"


class SpanColor(str, Enum):
    """
    The color of the span, this is optional and currently only used for NVTX spans.
    Here we list the colors that are supported by NVTX.
    """

    UNSET = "unset"
    GREEN = "green"
    BLUE = "blue"
    YELLOW = "yellow"
    PURPLE = "purple"
    RAPIDS = "rapids"
    CYAN = "cyan"
    RED = "red"
    WHITE = "white"
    DARK_GREEN = "darkgreen"
    ORANGE = "orange"


class Span:
    """
    A span records an operation that takes some time to complete. It is identified by a name,
    and has a duration and an optional color for backends that support colors such as NVTX.
    A span starts and ends with an event, and can have optional child events.
    """

    def __init__(
        self,
        id: uuid.UUID,
        name: SpanName,
        start_event: Event,
        start_time: Optional[float] = None,
        color: Optional[SpanColor] = None,
        verbosity: Verbosity = Verbosity.INFO,
    ):
        self.id: uuid.UUID = id
        self.name: SpanName = name
        self.duration: Duration = Duration.create(start_time=start_time, start=True)
        self.events: list[Event] = [start_event]
        self.metrics: EventMetrics = EventMetrics()
        self.color: Optional[SpanColor] = color
        self.verbosity: Verbosity = verbosity

    @classmethod
    def create(
        cls,
        name: SpanName,
        color: Optional[SpanColor] = None,
        start_time: Optional[float] = None,
        verbosity: Verbosity = Verbosity.INFO,
        metrics: Optional[EventMetrics] = None,
    ) -> "Span":
        """
        Create a new span with the given parameters.

        Args:
            name: The name of the span.
            start: The start event.
            color: The optional color of the span.
            start_time: The optional start time of the span, the current timestamp will be used if not specified.
            verbosity: The verbosity of the span, INFO by default.
            metrics: The optional metrics of the span, the start event will have these metrics added to it.
        """
        start_event = Event.create(name=EventName.SPAN_START, metrics=metrics)
        return cls(
            id=uuid.uuid4(), name=name, start_event=start_event, start_time=start_time, color=color, verbosity=verbosity
        )

    @property
    def start_event(self) -> Event:
        assert isinstance(self.events[0], Event)
        return self.events[0]

    @property
    def stop_event(self) -> Optional[Event]:
        return self.events[-1] if len(self.events) > 1 and not self.duration.running else None

    def stop(self) -> None:
        if not self.duration.running:
            return  # already stopped
        self.add_event(Event.create(name=EventName.SPAN_STOP, metrics=self.metrics))
        self.duration.stop(reset=False)

    def add_event(self, event: Event) -> None:
        self.events.append(event)

    def add_metric(self, name: str, value: float) -> EventMetric:
        """
        Add a metric to the span. This metric will be recorded with the end event.
        If you need to record a metric before the span ends, use the add_event() method,
        making sure the event has the metric you want to record.

        Args:
            name: The name of the metric.
            value: The value of the metric.

        Returns:
            The metric.
        """
        self.metrics[name] = EventMetric(name=name, value=value)
        return self.metrics[name]

    def add_metrics(self, metrics: EventMetrics) -> None:
        """
        Add a dictionary of metrics to the event.
        """
        self.metrics.update(metrics)



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
import os
from typing import Optional

import torch.distributed as dist


def get_rank(group: Optional[dist.ProcessGroup] = None) -> int:
    """Get the rank (GPU device) of the worker.

    Returns:
        rank (int): The rank of the worker.
    """
    rank = int(os.getenv("RANK", "0"))
    if dist.is_available() and dist.is_initialized():
        rank = dist.get_rank(group)
    return rank


def get_world_size(group: Optional[dist.ProcessGroup] = None) -> int:
    """Get world size. How many GPUs are available in this job.

    Returns:
        world_size (int): The total number of GPUs available in this job.
    """
    world_size = int(os.getenv("WORLD_SIZE", "1"))
    if dist.is_available() and dist.is_initialized():
        world_size = dist.get_world_size(group)
    return world_size


def is_rank0() -> bool:
    """Check if current process is the master GPU.

    Returns:
        (bool): True if this function is called from the master GPU, else False.
    """
    return get_rank() == 0


def barrier() -> None:
    """Barrier for all ranks."""
    if dist.is_available() and dist.is_initialized():
        dist.barrier()



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#

import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional


@dataclass
class ApplicationConfig:
    """Configuration for the application creating the telemetry data."""

    # The job name can be used to group jobs with the same training characteristics.
    job_name: str = "unknown"
    # The job id and the job_restart_count should uniquely identify a job run.
    job_id: str = "unknown"
    # The global rank of the process.
    rank: int = 0
    # The total number of ranks in the job run.
    world_size: int = 1
    # If the same job is restarted without changing the job id, this needs to be incremented.
    job_restart_count: int = 0
    # The name of the environment the job is running in, for example the cluster name, if known
    environment: str = "unknown"


class BackendType(str, Enum):
    """Supported backend types for telemetry data."""

    LOGGER = "logger"
    FILE = "file"
    NVTX = "nvtx"
    OTLP_TRACES = "otlp_traces"
    OTLP_LOGS = "otlp_logs"


@dataclass
class BackendConfig:
    """Configuration for a specific backend."""

    # The backend type, we currently support logger, file, and Open Telemetry (OTEL).
    backend_type: BackendType


@dataclass
class LoggerBackendConfig(BackendConfig):
    """Configuration for the logger backend, which logs JSON events using a python logger."""

    backend_type: BackendType = BackendType.LOGGER

    # The name of the logger to use.
    name: str = "training_telemetry"
    # If this is true, then the log context will be propagated to the root logger,
    # which could result in duplicate logs, so normally this should be false.
    # However, unit tests need this to be true to capture the log messages for testing
    propagate_log_context: bool = False
    # The format of the log message.
    log_format: Optional[str] = None
    # The log level to use for the logger.
    log_level: int = logging.INFO
    # If rank_aware is true, then the rank number will be added to the logger name.
    # For example, if the logger name is "training_telemetry" and the rank is 0,
    # then the logger name will be "training_telemetry_rank0".
    rank_aware: bool = True
    # If this is true, then only error events will be logged.
    errors_only: bool = False


@dataclass
class FileBackendConfig(BackendConfig):
    """Configuration for the file backend, which logs JSON events to a file."""

    backend_type: BackendType = BackendType.FILE
    # The output file path without the extension, should contain the directory
    # and the base name of the file.
    # The extension will be added automatically as .json, the only
    # format currently supported by the file backend. Later, we may
    # add .csv or other formats.
    output_file_path: str = "/tmp/training_telemetry"
    # If this is true, then the file path will contain the rank number
    # before the .json extension. For example, if the output file path
    # is "/tmp/training_telemetry" and the rank is 0, then the file will be "/tmp/training_telemetry_rank0.json".
    # The backend will likely be created before the torch distributed process group is initialized,
    # so we need to use the RANK environment variable to determine the rank number.
    rank_aware: bool = True
    # If this is true, then the file path will contain the date and time
    # after the rank number. For example, if the output file path
    # is "/tmp/training_telemetry" and the date and time is 2021-01-01 12:00:00, then the file
    # will be "/tmp/training_telemetry_rank0_20210101_120000.json".
    date_aware: bool = True


@dataclass
class OTLPTracesBackendConfig(BackendConfig):
    """
    Configuration for the Open Telemetry traces backend,
    which converts spans and events to OTEL traces.
    """

    backend_type: BackendType = BackendType.OTLP_TRACES
    # The exporter to use, OTLP can be send over grpc or http, and console will print to the console.
    exporter: str = "grpc"  # or "http", "console"
    # The endpoint when using either grpc or http as an OTLP exporter, the grpc port is normally 4317 and the http port is 4318.
    endpoint: str = "http://localhost:4317"  # or "http://localhost:4318"


class OTLPLogsFormat(str, Enum):
    """The format of the logs to send over OTLP."""

    # This format is identical to the python logger backend, which logs event
    # properties as key = value pairs separated by a pipe character.
    KEY_VALUE = "key_value"
    # This format is identical to the file backend, which logs events as JSON.
    JSON = "json"


@dataclass
class OTLPLogsBackendConfig(BackendConfig):
    """
    Configuration for the Open Telemetry logs backend. This backend
    encodes events as either key=value pairs (similarly to the logger backend)
    or JSON (similarly to the file backend), and sends these strings as logs over
    OTLP to an OTLP logs endpoint.
    """

    backend_type: BackendType = BackendType.OTLP_LOGS
    # The name of the logger to use.
    name: str = "otlp_logs"
    # The log level to use for the logger.
    log_level: int = logging.INFO
    # If rank_aware is true, then the rank number will be added to the logger name.
    # For example, if the logger name is "training_telemetry" and the rank is 0,
    # then the logger name will be "training_telemetry_rank0".
    rank_aware: bool = True
    # The exporter to use, OTLP can be send over grpc or http, and console will print to the console.
    exporter: str = "grpc"  # or "http", "console"
    # The endpoint when using either grpc or http as an OTLP exporter, the grpc port is normally 4317 and the http port is 4318.
    endpoint: str = "http://localhost:4317"  # or "http://localhost:4318"
    # The format of the logs to send over OTLP.
    format: OTLPLogsFormat = OTLPLogsFormat.KEY_VALUE


@dataclass
class NVTXBackendConfig(BackendConfig):
    """Configuration for the NVTX backend, which logs JSON events to a file."""

    backend_type: BackendType = BackendType.NVTX


def default_backends() -> list[BackendConfig]:
    """Return the default backends, used if the user doesn't provide any backends."""
    return [
        LoggerBackendConfig(),
        FileBackendConfig(),
        NVTXBackendConfig(),
    ]


@dataclass
class TelemetryConfig:
    """Configuration for the telemetry system."""

    # The backends configuration is used to determine which backends are enabled, and their configuration.
    backends: List[BackendConfig] = field(default_factory=default_backends)

    # The application configuration contains information that might be used by backends to set application properties.
    application: ApplicationConfig = field(default_factory=ApplicationConfig)

    # The log level and format used by the library to log any initialization or error messages.
    # Set to CRITICAL to disable logging completely as the highest level used is ERROR.
    log_level: int = logging.INFO
    log_format: Optional[str] = None



#
#  Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.
#
# NVIDIA CORPORATION and its licensors retain all intellectual property
# and proprietary rights in and to this software, related documentation
# and any modifications thereto.  Any use, reproduction, disclosure or
# distribution of this software and related documentation without an
# express license agreement from NVIDIA CORPORATION is strictly
# prohibited.
#
from enum import Enum


class Verbosity(int, Enum):
    """
    The verbosity level associated with a backend. They go from from smallest to highest verbosity.
    A backend will record all events and spans up to a certain verbosity level.
    """

    # A verbosity intended for non rank-0 logging backends, where we only want to log errors.
    ERROR = 0
    # The default verbosity for logging and file backends, normally on rank 0 for logging.
    INFO = 1
    # A backend that supports optimized tracing, such as the OTEL protocol or compressed files, e.g. protobufs.
    TRACING = 3
    # A backend that supports profiling, such as the NVTX backend for the Nsight Systems profiler.
    PROFILING = 4
